[PATCH v2] libcamera: controls: Add overwriteExisting parameter to ControlList::merge

Stefan Klug stefan.klug at ideasonboard.com
Mon Mar 4 16:58:53 CET 2024


This is useful in many cases although not included in the stl.

Signed-off-by: Stefan Klug <stefan.klug at ideasonboard.com>
---

This version adds a unittest and fixes a doxygen doc.

 include/libcamera/controls.h   |  2 +-
 src/libcamera/controls.cpp     |  9 ++++--
 test/controls/control_list.cpp | 50 ++++++++++++++++++++++++++++++++++
 3 files changed, 57 insertions(+), 4 deletions(-)

diff --git a/include/libcamera/controls.h b/include/libcamera/controls.h
index cf942055..0bf778d8 100644
--- a/include/libcamera/controls.h
+++ b/include/libcamera/controls.h
@@ -368,7 +368,7 @@ public:
 	std::size_t size() const { return controls_.size(); }
 
 	void clear() { controls_.clear(); }
-	void merge(const ControlList &source);
+	void merge(const ControlList &source, bool overwriteExisting = false);
 
 	bool contains(unsigned int id) const;
 
diff --git a/test/controls/control_list.cpp b/test/controls/control_list.cpp
index c03f230e..304cabfc 100644
--- a/test/controls/control_list.cpp
+++ b/test/controls/control_list.cpp
@@ -196,6 +196,56 @@ protected:
 			return TestFail;
 		}
 
+		/*
+		 * Create two lists with overlapping controls. Merge them with
+		 * overwriteExisting = true, verifying that the existing control
+		 * values *get* overwritten.
+		 */
+		mergeList.clear();
+		mergeList.set(controls::Brightness, 0.7f);
+		mergeList.set(controls::Saturation, 0.4f);
+
+		list.clear();
+		list.set(controls::Brightness, 0.5f);
+		list.set(controls::Contrast, 1.1f);
+
+		mergeList.merge(list, true);
+		if (mergeList.size() != 3) {
+			cout << "Merged list should contain three elements" << endl;
+			return TestFail;
+		}
+
+		if (list.size() != 2) {
+			cout << "The list to merge should contain two elements"
+			     << endl;
+			return TestFail;
+		}
+
+		if (!mergeList.get(controls::Brightness) ||
+		    !mergeList.get(controls::Contrast) ||
+		    !mergeList.get(controls::Saturation)) {
+			cout << "Merged list does not contain all controls" << endl;
+			return TestFail;
+		}
+
+		if (mergeList.get(controls::Brightness) != 0.5f) {
+			cout << "Brightness control value did not change after merging lists"
+			     << endl;
+			return TestFail;
+		}
+
+		if (mergeList.get(controls::Contrast) != 1.1f) {
+			cout << "Contrast control value did not change after merging lists"
+			     << endl;
+			return TestFail;
+		}
+
+		if (mergeList.get(controls::Saturation) != 0.4f) {
+			cout << "Saturation control value changed after merging lists"
+			     << endl;
+			return TestFail;
+		}
+
 		return TestPass;
 	}
 };
diff --git a/src/libcamera/controls.cpp b/src/libcamera/controls.cpp
index b808116c..93ce4efe 100644
--- a/src/libcamera/controls.cpp
+++ b/src/libcamera/controls.cpp
@@ -910,10 +910,13 @@ ControlList::ControlList(const ControlInfoMap &infoMap,
 /**
  * \brief Merge the \a source into the ControlList
  * \param[in] source The ControlList to merge into this object
+ * \param[in] overwriteExisting Control if existing elements in *this shall be
+ * overwritten
  *
  * Merging two control lists copies elements from the \a source and inserts
  * them in *this. If the \a source contains elements whose key is already
  * present in *this, then those elements are not overwritten.
  *
  * Only control lists created from the same ControlIdMap or ControlInfoMap may
  * be merged. Attempting to do otherwise results in undefined behaviour.
@@ -921,7 +924,7 @@ ControlList::ControlList(const ControlInfoMap &infoMap,
  * \todo Reimplement or implement an overloaded version which internally uses
  * std::unordered_map::merge() and accepts a non-const argument.
  */
-void ControlList::merge(const ControlList &source)
+void ControlList::merge(const ControlList &source, bool overwriteExisting)
 {
 	/**
 	 * \todo ASSERT that the current and source ControlList are derived
@@ -936,7 +939,7 @@ void ControlList::merge(const ControlList &source)
 	 */
 
 	for (const auto &ctrl : source) {
-		if (contains(ctrl.first)) {
+		if (!overwriteExisting && contains(ctrl.first)) {
 			const ControlId *id = idmap_->at(ctrl.first);
 			LOG(Controls, Warning)
 				<< "Control " << id->name() << " not overwritten";
@@ -910,10 +910,13 @@ ControlList::ControlList(const ControlInfoMap &infoMap,
  * Merging two control lists copies elements from the \a source and inserts
  * them in *this. If the \a source contains elements whose key is already
- * present in *this, then those elements are not overwritten.
+ * present in *this, then those elements are only overwritten if
+ * \a overwriteExisting is true.
  *
  * Only control lists created from the same ControlIdMap or ControlInfoMap may
-- 
2.40.1



More information about the libcamera-devel mailing list